-------------------------------------------------------------------------------
--
-- Chess Board VGA controller for the Digilent Spartan 3 starter board
--
-- Uses an OPB interface, e.g., for use with the Microblaze soft core
--
-- Sheng-Kuo Lu
-- sklu.iic93g@nctu.edu.tw
--
-- Update from the sample Text Mode VGA controller
-- Stephen A. Edwards
-- sedwards@cs.columbia.edu
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity opb_s3board_vga is

  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(31 downto 0) := X"FEFF1000";
    C_HIGHADDR   : std_logic_vector(31 downto 0) := X"FEFF1FFF");

  port (
    OPB_Clk        : in std_logic;
    OPB_Rst        : in std_logic;

    -- OPB signals
    OPB_ABus       : in std_logic_vector (31 downto 0);
    OPB_BE         : in std_logic_vector (3 downto 0);
    OPB_DBus       : in std_logic_vector (31 downto 0);
    OPB_RNW        : in std_logic;
    OPB_select     : in std_logic;
    OPB_seqAddr    : in std_logic;

    VGA_DBus       : out std_logic_vector (31 downto 0);
    VGA_errAck     : out std_logic;
    VGA_retry      : out std_logic;
    VGA_toutSup    : out std_logic;
    VGA_xferAck    : out std_logic;

    Pixel_Clock_2x : in std_logic;

    -- Video signals
    VIDOUT_CLK     : out std_logic;
    VIDOUT_RED     : out std_logic;
    VIDOUT_GREEN   : out std_logic;
    VIDOUT_BLUE    : out std_logic;
    VIDOUT_HSYNC : out std_logic;
    VIDOUT_VSYNC : out std_logic);

end opb_s3board_vga;

architecture Behavioral of opb_s3board_vga is

  constant BASEADDR : std_logic_vector(31 downto 0) := X"FEFF1000";

  -- Video parameters

  constant HTOTAL : integer := 800;
  constant HSYNC : integer := 96;
  constant HBACK_PORCH : integer := 48;
  constant HACTIVE : integer := 640;
  constant HFRONT_PORCH : integer := 16;

  constant VTOTAL : integer := 525;
  constant VSYNC : integer := 2;
  constant VBACK_PORCH : integer := 33;
  constant VACTIVE : integer := 480;
  constant VFRONT_PORCH : integer := 10;

  -- 2048 X 9 dual-ported Xilinx block RAM
  component RAMB16_S9_S9
    port (
      DOA   : out std_logic_vector (7 downto 0);
      DOPA0  : out std_logic;
      ADDRA : in std_logic_vector (10 downto 0);
      CLKA  : in std_logic;
      DIA   : in std_logic_vector (7 downto 0);
      DIPA0  : in std_logic;
      ENA   : in std_logic;
      SSRA  : in std_logic;
      WEA   : in std_logic;
      DOB   : out std_logic_vector (7 downto 0);
      DOPB0  : out std_logic;
      ADDRB : in std_logic_vector (10 downto 0);
      CLKB  : in std_logic;
      DIB   : in std_logic_vector (7 downto 0);
      DIPB0  : in std_logic;
      ENB   : in std_logic;
      SSRB  : in std_logic;
      WEB   : in std_logic);
  end component;

  -- 512 X 36 dual-ported Xilinx block RAM
  component RAMB16_S36_S36
    port (
      DOA   : out std_logic_vector (31 downto 0);
      DOPA  : out std_logic_vector(3 downto 0);
      ADDRA : in std_logic_vector (8 downto 0);
      CLKA  : in std_logic;
      DIA   : in std_logic_vector (31 downto 0);
      DIPA  : in std_logic_vector(3 downto 0);
      ENA   : in std_logic;
      SSRA  : in std_logic;
      WEA   : in std_logic;
      DOB   : out std_logic_vector (31 downto 0);
      DOPB  : out std_logic_vector(3 downto 0);
      ADDRB : in std_logic_vector (8 downto 0);
      CLKB  : in std_logic;
      DIB   : in std_logic_vector (31 downto 0);
      DIPB  : in std_logic_vector(3 downto 0);
      ENB   : in std_logic;
      SSRB  : in std_logic;
      WEB   : in std_logic);
  end component;

  -- Signals latched from the OPB
  signal ABus : std_logic_vector (31 downto 0);
  signal DBus : std_logic_vector (31 downto 0);
  signal RNW  : std_logic;
  signal select_delayed : std_logic;

  -- Latched output signals for the OPB
  signal DBus_out : std_logic_vector (31 downto 0);

  -- Signals for the OPB-mapped RAM
  signal ChipSelect : std_logic;            -- Address decode
  signal MemCycle1, MemCycle2 : std_logic;  -- State bits
  signal RamPageAddress : std_logic;
  signal RamSelect : std_logic_vector (1 downto 0);
  signal RST, WE : std_logic_vector(1 downto 0);
  signal DOUT0 : std_logic_vector(7 downto 0);
  signal DOUT1 : std_logic_vector(31 downto 0);
  signal ReadData : std_logic_vector(31 downto 0);

  -- Signals for the video controller
  signal Pixel_Clock : std_logic;       -- 25 MHz clock divided from 50 MHz
  signal LoadNShift : std_logic;        -- Shift register control
  signal PieceData : std_logic_vector(31 downto 0);   -- Input to shift register
  signal ShiftData : std_logic_vector(31 downto 0);  -- Shift register data
  signal VideoData : std_logic;         -- Serial out ANDed with blanking

  signal Hcount : std_logic_vector(9 downto 0);  -- Horizontal position (0-800)
  signal Vcount : std_logic_vector(9 downto 0);  -- Vertical position (0-524)
  signal HBLANK_N, VBLANK_N : std_logic;   -- Blanking signals

  signal PieceLoad, LoadAddr : std_logic;  -- Font/Character RAM read triggers
  signal PieceAddr : std_logic_vector(8 downto 0);
  signal BlockAddr : std_logic_vector(10 downto 0);
  signal BlockColumn : std_logic_vector(9 downto 0);
  signal BlockRow : std_logic_vector(9 downto 0);
  signal Column : std_logic_vector(4 downto 0); -- 0-19
  signal Row : std_logic_vector(3 downto 0);   -- 0-14
  signal EndOfLine, EndOfField : std_logic;
  signal Player, Cursor : std_logic;
  signal PlayerData, CursorData : std_logic;
  signal Background, BackgroundData : std_logic_vector(1 downto 0);

  signal DOUTB0 : std_logic_vector(7 downto 0);
  signal DOUTB1 : std_logic_vector(31 downto 0);
  signal DOP0A, DOP0B : std_logic;
  signal DOP1A, DOP1B : std_logic_vector(3 downto 0);

  attribute INIT_00 : string;
  attribute INIT_01 : string;
  attribute INIT_02 : string;
  attribute INIT_03 : string;
  attribute INIT_04 : string;
  attribute INIT_05 : string;
  attribute INIT_06 : string;
  attribute INIT_07 : string;
  attribute INIT_08 : string;
  attribute INIT_09 : string;
  attribute INIT_0a : string;
  attribute INIT_0b : string;
  attribute INIT_0c : string;
  attribute INIT_0d : string;
  attribute INIT_0e : string;
  attribute INIT_0f : string;
  attribute INIT_10 : string;
  attribute INIT_11 : string;
  attribute INIT_12 : string;
  attribute INIT_13 : string;
  attribute INIT_14 : string;
  attribute INIT_15 : string;
  attribute INIT_16 : string;
  attribute INIT_17 : string;
  attribute INIT_18 : string;
  attribute INIT_19 : string;
  attribute INIT_1a : string;
  attribute INIT_1b : string;
  attribute INIT_1c : string;
  attribute INIT_1d : string;
  attribute INIT_1e : string;
  attribute INIT_1f : string;
  attribute INIT_20 : string;
  attribute INIT_21 : string;
  attribute INIT_22 : string;
  attribute INIT_23 : string;
  attribute INIT_24 : string;
  attribute INIT_25 : string;
  attribute INIT_26 : string;
  attribute INIT_27 : string;
  attribute INIT_28 : string;
  attribute INIT_29 : string;
  attribute INIT_2a : string;
  attribute INIT_2b : string;
  attribute INIT_2c : string;
  attribute INIT_2d : string;
  attribute INIT_2e : string;
  attribute INIT_2f : string;
  attribute INIT_30 : string;
  attribute INIT_31 : string;
  attribute INIT_32 : string;
  attribute INIT_33 : string;
  attribute INIT_34 : string;
  attribute INIT_35 : string;
  attribute INIT_36 : string;
  attribute INIT_37 : string;
  attribute INIT_38 : string;
  attribute INIT_39 : string;
  attribute INIT_3a : string;
  attribute INIT_3b : string;
  attribute INIT_3c : string;
  attribute INIT_3d : string;
  attribute INIT_3e : string;
  attribute INIT_3f : string;

  -- Initial chess piece data
  attribute INIT_00 of Board_Block : label is
   "0000000000000000000000000000000000000000000000000000000000000007";
  attribute INIT_01 of Board_Block : label is
   "0000000000000000000000000000000000000000000000000000000000000000";
  attribute INIT_02 of Board_Block : label is
   "0000050403010203040500000000000000000000000000000000000000000000";
  attribute INIT_03 of Board_Block : label is
   "4000000000000000000000000000060606060606060600000000000000000000";
  attribute INIT_04 of Board_Block : label is
   "0000000000000000000000000000000000000000000000000000000000000000";
  attribute INIT_05 of Board_Block : label is
   "0000000000000000000000000000000000000000000000000000000000000000";
  attribute INIT_06 of Board_Block : label is
   "0000000000000000000026262626262626260000000000000000000000000000";
  attribute INIT_07 of Board_Block : label is
   "0000000000000000000000000000000000000000000025242321222324250000";


  -- Standard 8x16 font taken from the Linux console font file "lat0-16.psfu"

  attribute INIT_00 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";  -- empty piece
  attribute INIT_01 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";  -- empty piece
  attribute INIT_02 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";  -- empty piece
  attribute INIT_03 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- empty piece
  attribute INIT_04 of Piece_Bitmap : label is
   "0000800000008000000080000003E00000008000000080000000000000000000";  -- the KING
  attribute INIT_05 of Piece_Bitmap : label is
   "0BFD5FE80BFD5FE803FB6FE805E6B3D0061EBC3001F2A7C0000220000001C000";  -- the KING
  attribute INIT_06 of Piece_Bitmap : label is
   "004FF9000060030000FFFF80008FF88001A082C0037EBF6006FEBFA005FEBFD0";  -- the KING
  attribute INIT_07 of Piece_Bitmap : label is
   "000000000000000000000000000FF800007FFF0000FFFF800087F08000700700";	-- the KING
  attribute INIT_08 of Piece_Bitmap : label is
   "3C6186381861861800F38E0000F3EE000061C600000380000001000000000000";  -- the QUEEN
  attribute INIT_09 of Piece_Bitmap : label is
   "07FFFFE007FFFEE00679FEF00F71CE300C73CE301E73CE381863CE183C638E38";  -- the QUEEN
  attribute INIT_0a of Piece_Bitmap : label is
   "009FF80000C0038001FFFF800180018001FFFF8003FFFF8007FFFF8007FFFFE0";  -- the QUEEN
  attribute INIT_0b of Piece_Bitmap : label is
   "0000000000000000000000000000000000FFFE0001FFFF00018FF00001E00600";	-- the QUEEN
  attribute INIT_0c of Piece_Bitmap : label is
   "0001C0000003E0000003E0000003E0000001C000000000000000000000000000";	-- the BUSHOP
  attribute INIT_0d of Piece_Bitmap : label is
   "007F7E00007F7F00007C1F00007F7F00003F7E00001FFE00001FFC000007F000";	-- the BUSHOP
  attribute INIT_0e of Piece_Bitmap : label is
   "001FFC000017F40000080800000FF800000FF80000180C00003FFE00003FFE00";	-- the BUSHOP
  attribute INIT_0f of Piece_Bitmap : label is
   "00000000000000000400001007FE3FF00FFF7FF80FFFFFF80006B0000007F000";	-- the BUSHOP
  attribute INIT_10 of Piece_Bitmap : label is
   "007FF800007FE0000077000000C3000000020000000000000000000000000000";	-- the KNIGHT
  attribute INIT_11 of Piece_Bitmap : label is
   "03FF7F4003FFBE4003FFBEC003FFBEC001BFFC8001DFF30000FFE60000E78C00";	-- the KNIGHT
  attribute INIT_12 of Piece_Bitmap : label is
   "0183FF980F83FF980EC1FF901FE1FFB01BF8FF300FFEFF300FFFFF3007FFFF20";	-- the KNIGHT
  attribute INIT_13 of Piece_Bitmap : label is
   "0000000000000000001FFFF8001FFFF8001FFF98001FFF98000FFF98000FFF98";	-- the KNIGHT
  attribute INIT_14 of Piece_Bitmap : label is
   "01FFFF0001FFFF0001E7CF0001E7CF0000000000000000000000000000000000";	-- the ROOK
  attribute INIT_15 of Piece_Bitmap : label is
   "003FF800003FF800003FF800003FF80000600C0000FFFE000180030001FFFF00";	-- the ROOK
  attribute INIT_16 of Piece_Bitmap : label is
   "01FFFF000180030000FFFE0000600C00003FF800003FF800003FF800003FF800";	-- the ROOK
  attribute INIT_17 of Piece_Bitmap : label is
   "00000000000000000000000007FFFFC007FFFFC007FFFFC0060000C001FFFF00";	-- the ROOK
  attribute INIT_18 of Piece_Bitmap : label is
   "0007F0000003F0000003E0000001C00000000000000000000000000000000000";	-- the PAWN
  attribute INIT_19 of Piece_Bitmap : label is
   "000FF800000FFC00001FFC00001FFC00001FFC00000FF8000007F0000003E000";	-- the PAWN
  attribute INIT_1a of Piece_Bitmap : label is
   "00FFFF8000FFFF8000FFFF80007FFF80007FFF00003FFE00001FFC000007F000";	-- the PAWN
  attribute INIT_1b of Piece_Bitmap : label is
   "000000000000000000000000000000000000000000FFFF8000FFFF8000FFFF80";	-- the PAWN
  attribute INIT_1c of Piece_Bitmap : label is
   "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF";	-- full foreground
  attribute INIT_1d of Piece_Bitmap : label is
   "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF";	-- full foreground
  attribute INIT_1e of Piece_Bitmap : label is
   "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF";	-- full foreground
  attribute INIT_1f of Piece_Bitmap : label is
   "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF";	-- full foreground
  attribute INIT_20 of Piece_Bitmap : label is
   "F000000FF000000FF000000FF000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF";	-- selected empty piece
  attribute INIT_21 of Piece_Bitmap : label is
   "F000000FF000000FF000000FF000000FF000000FF000000FF000000FF000000F";	-- selected empty piece
  attribute INIT_22 of Piece_Bitmap : label is
   "F000000FF000000FF000000FF000000FF000000FF000000FF000000FF000000F";	-- selected empty piece
  attribute INIT_23 of Piece_Bitmap : label is
   "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000FF000000FF000000FF000000F";	-- selected empty piece
  attribute INIT_24 of Piece_Bitmap : label is
   "800380018001000180010001800100018007C0018001000180010001FFFFFFFF";	-- selected KING
  attribute INIT_25 of Piece_Bitmap : label is
   "8801002190000011900280199006C001980C6031843C784185E44701800C6001";	-- selected KING
  attribute INIT_26 of Piece_Bitmap : label is
   "809FF20180E0060180BFFB018180030181C00701833FF981860100C184010061";	-- selected KING
  attribute INIT_27 of Piece_Bitmap : label is
   "FFFFFFFF800000018000000180000001800FE00180F01E018180030181E00F01";	-- selected KING
  attribute INIT_28 of Piece_Bitmap : label is
   "9230C3138C30C30D8048C48180492481803123018000C00180000001FFFFFFFF";	-- selected QUEEN
  attribute INIT_29 of Piece_Bitmap : label is
   "82E211D182A619518325293987292D39862925198429250D8C31C30D9230C313";	-- selected QUEEN
  attribute INIT_2a of Piece_Bitmap : label is
   "80700181804FFE8180C00081809FFE418180006181E001E1813FFF21836291A1";	-- selected QUEEN
  attribute INIT_2b of Piece_Bitmap : label is
   "FFFFFFFF80000001800000018007FC018078038180C000C180F007C1804FFC81";	-- selected QUEEN
  attribute INIT_2c of Piece_Bitmap : label is
   "800EE001800380018006C001800440018006C0018003800180000001FFFFFFFF";	-- selected BUSHOP
  attribute INIT_2d of Piece_Bitmap : label is
   "8040040180C10601808102018087C201804104018041040180200C0180383801";	-- selected BUSHOP
  attribute INIT_2e of Piece_Bitmap : label is
   "800FE00180383801803838018017D0018010100180183001803FF80180200C01";	-- selected BUSHOP
  attribute INIT_2f of Piece_Bitmap : label is
   "FFFFFFFF8000000180000001880000218FFC7FE19006C0119FF39FF1800D6001";	-- selected BUSHOP
  attribute INIT_30 of Piece_Bitmap : label is
   "80D08C01805CF801805DE0018075000180C300018002000180000001FFFFFFFF";	-- selected KNIGHT
  attribute INIT_31 of Piece_Bitmap : label is
   "840080618600C061820040C181000081814000818160018181A0030180800601";	-- selected KNIGHT
  attribute INIT_32 of Piece_Bitmap : label is
   "80040019818600198F8100198D410019986180119A308031880E80318C038031";	-- selected KNIGHT
  attribute INIT_33 of Piece_Bitmap : label is
   "FFFFFFFF8000000180000001801FFFF980100019801000198010001980080019";	-- selected KNIGHT
  attribute INIT_34 of Piece_Bitmap : label is
   "8100010181000101813C79018124490181E7CF018000000180000001FFFFFFFF";	-- selected ROOK
  attribute INIT_35 of Piece_Bitmap : label is
   "8020080180200801802008018020080180200801807FFC0180C0060181FFFF01";	-- selected ROOK
  attribute INIT_36 of Piece_Bitmap : label is
   "810001018100010181FFFF0180C00601807FFC01802008018020080180200801";	-- selected ROOK
  attribute INIT_37 of Piece_Bitmap : label is
   "FFFFFFFF80000001800000018000000187FFFFC1840000418400004187FFFFC1";	-- selected ROOK
  attribute INIT_38 of Piece_Bitmap : label is
   "80063001800630018003600180018001800000018000000180000001FFFFFFFF";	-- selected PAWN
  attribute INIT_39 of Piece_Bitmap : label is
   "800C180180180C01801004018010040180080C01800C18018006300180022001";	-- selected PAWN
  attribute INIT_3a of Piece_Bitmap : label is
   "808000818080008180C00181804001018020030180300601801C1C0180063001";	-- selected PAWN
  attribute INIT_3b of Piece_Bitmap : label is
   "FFFFFFFF8000000180000001800000018000000180FFFF818080008180800081";	-- selected PAWN
  attribute INIT_3c of Piece_Bitmap : label is
   "CDBFFFFFCDBFFFFFDDBF3FFFDDBF3FFF889F3FFFFFFFFFFFFFFFFFFFFFFFFFFF";	-- win
  attribute INIT_3d of Piece_Bitmap : label is
   "E87F3F79E87F3E3BEC7E3C33ECBF3F43CCBFFFFFCDBFFFFFCDBFFFFFCDBFFFFF";	-- win
  attribute INIT_3e of Piece_Bitmap : label is
   "E67F3F79E67F3F79E67F3F79E67F3F79E27F3F79EA7F3F79EA7F3F79E87F3F79";	-- win
  attribute INIT_3f of Piece_Bitmap : label is
   "FFFFFFFFFFFFFFFFF6FC1C30F6FF3E79F6FF3F79F6FF3F79F6FF3F79F67F3F79";	-- win

begin  -- Behavioral

  -----------------------------------------------------------------------------
  --
  -- Instances of the block RAMs
  -- Each is 2K bytes, so 4K total
  --
  -----------------------------------------------------------------------------

  -- First 2K is used for Pieces    (only 300 bytes in use)
  -- Remaining 2K holds the piece bitmap  (1 block RAMs)
  --

  -- Port A is used for communication with the OPB
  -- Port B is for video

  Board_Block : RAMB16_S9_S9
    port map (
      DOA   => DOUT0,
      DOPA0  => DOP0A,
      ADDRA => ABus(10 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(7 downto 0),
      DIPA0  => '1',
      ENA   => '1',
      SSRA  => RST(0),
      WEA   => WE(0),
      DOB   => DOUTB0,
      DOPB0  => DOP0B,
      ADDRB => BlockAddr(10 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00",
      DIPB0  => '1',
      ENB   => '1',
      SSRB  => '0',
      WEB   => '0');

    Piece_Bitmap : RAMB16_S36_S36
    port map (
      DOA   => DOUT1,
      DOPA  => DOP1A,
      ADDRA => ABus(8 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(31 downto 0),
      DIPA  => "1111",
      ENA   => '1',
      SSRA  => RST(1),
      WEA   => WE(1),
      DOB   => DOUTB1,
      DOPB  => DOP1B,
      ADDRB => PieceAddr(8 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00000000",
      DIPB  => "1111",
      ENB   => '1',
      SSRB  => '0',
      WEB   => '0');

  -----------------------------------------------------------------------------
  --
  -- OPB-RAM controller
  --
  -----------------------------------------------------------------------------

  -- Unused OPB control signals
  VGA_errAck <= '0';
  VGA_retry <= '0';
  VGA_toutSup <= '0';

  -- Latch the relevant OPB signals from the OPB, since they arrive late
  LatchOPB: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Abus <= ( others => '0' );
      DBus <= ( others => '0' );
      RNW <= '1';
      select_delayed <= '0';
    elsif OPB_Clk'event and OPB_Clk = '1' then
      ABus <= OPB_ABus;
      DBus <= OPB_DBus;
      RNW <= OPB_RNW;
      select_delayed <= OPB_Select;
    end if;
  end process LatchOPB;

  -- Address bits 31 downto 12 is our chip select
  -- 11 is the RAM page select
  -- 10 downto 0 is the RAM byte select

  ChipSelect <=
    '1' when select_delayed = '1' and
       (ABus(31 downto 12) = BASEADDR(31 downto 12)) and
       MemCycle1 = '0' and MemCycle2 = '0' else
    '0';

  RamPageAddress <= ABus(11);

  RamSelect <=
    "01" when RamPageAddress = '0' else
    "10" when RamPageAddress = '1' else
    "00";

  MemCycleFSM : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      MemCycle1 <= '0';
      MemCycle2 <= '0';
    elsif OPB_Clk'event and OPB_Clk = '1' then
      MemCycle2 <= MemCycle1;
      MemCycle1 <= ChipSelect;
    end if;
  end process MemCycleFSM;

  VGA_xferAck <= MemCycle2;

  WE <= RamSelect when ChipSelect = '1' and RNW = '0' and OPB_Rst = '0' else "00";

  RST <= not RamSelect when ChipSelect = '1' and RNW = '1' and OPB_Rst = '0' else "11";

  ReadData <= (DOUT0 & DOUT0 & DOUT0 & DOUT0) or DOUT1 when MemCycle1 = '1' else X"00000000";

  -- DBus(31 downto 24) is the byte for addresses ending in 0

  GenDOut: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      DBus_out <= ( others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
      DBus_out <= ReadData;
    end if;
  end process GenDOut;

  VGA_DBus <= DBus_out;

  ----------------------------------------------------------------------------
  --
  -- Video controller
  --
  -----------------------------------------------------------------------------

  -- Pixel clock divider

  Pixel_clk_divider : process (Pixel_Clock_2x, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Pixel_Clock <= '0';
    elsif Pixel_Clock_2x'event and Pixel_Clock_2x = '1' then
      Pixel_Clock <= not Pixel_Clock;
    end if;
  end process Pixel_clk_divider;

  -- Horizontal and vertical counters

  HCounter : process (Pixel_Clock, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Hcount <= (others => '0');
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine = '1' then
        Hcount <= (others => '0');
      else
        Hcount <= Hcount + 1;
      end if;
    end if;
  end process HCounter;

  EndOfLine <= '1' when Hcount = HTOTAL - 1 else '0';

  VCounter: process (Pixel_Clock, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Vcount <= (others => '0');
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine = '1' then
        if EndOfField = '1' then
          Vcount <= (others => '0');
        else
          Vcount <= Vcount + 1;
        end if;
      end if;
    end if;
  end process VCounter;

  EndOfField <= '1' when Vcount = VTOTAL - 1 else '0';

  -- State machines to generate HSYNC, VSYNC, HBLANK, and VBLANK

  HSyncGen : process (Pixel_Clock, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      VIDOUT_HSYNC <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine = '1' then
        VIDOUT_HSYNC <= '1';
      elsif Hcount = HSYNC - 1 then
        VIDOUT_HSYNC <= '0';
      end if;
    end if;
  end process HSyncGen;

  HBlankGen : process (Pixel_Clock, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      HBLANK_N <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if Hcount = HSYNC + HBACK_PORCH - 1 then
        HBLANK_N <= '1';
      elsif Hcount = HSYNC + HBACK_PORCH + HACTIVE - 1 then
        HBLANK_N <= '0';
      end if;
    end if;
  end process HBlankGen;

  VSyncGen : process (Pixel_Clock, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      VIDOUT_VSYNC <= '1';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine ='1' then
        if EndOfField = '1' then
          VIDOUT_VSYNC <= '1';
        elsif VCount = VSYNC - 1 then
          VIDOUT_VSYNC <= '0';
        end if;
      end if;
    end if;
  end process VSyncGen;

  VBlankGen : process (Pixel_Clock, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      VBLANK_N <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine ='1' then
        if Vcount = VSYNC + VBACK_PORCH - 1 then
          VBLANK_N <= '1';
        elsif VCount = VSYNC + VBACK_PORCH + VACTIVE - 1 then
          VBLANK_N <= '0';
        end if;
      end if;
    end if;
  end process VBlankGen;

  -- RAM read triggers and shift register control

  LoadNShift <= '1' when Hcount(4 downto 0) = X"0F" else '0';

  -- Correction of 4 needed to calculate the character address before the
  -- character is displayed
  BlockColumn <= Hcount - HSYNC - HBACK_PORCH + 32;
  Column <= BlockColumn(9 downto 5);  -- / 32
  BlockRow <= Vcount - VSYNC - VBACK_PORCH;
  Row <= BlockRow(8 downto 5);  -- / 32

  -- Column + Row * 20
  BlockAddr <= Column +
              ("000" & Row(3 downto 0) & "0000") +
              ("00000" & Row(3 downto 0) & "00");

  -- Most significant bit of character ignored

  PieceAddr(8 downto 5) <= DOUTB0(3 downto 0);
  PieceAddr(4 downto 0) <= BlockRow(4 downto 0);

  PieceData <= DOUTB1;
  PlayerData <= DOUTB0(5);
  CursorData <= DOUTB0(4);

  Board_background: process (Column, Row)
  begin
    if (Column > 5 and Column < 14) and (Row > 3 and Row < 12) then
      if Column(0) = Row(0) then
        BackgroundData <= "01";
      else
        BackgroundData <= "10";
      end if;
    else
      BackgroundData <= "00";
    end if;
  end process Board_background;

  -- Shift register

  ShiftRegister: process (Pixel_Clock, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      ShiftData <= X"00000000";
      Player <= '0';
      Cursor <= '0';
      Background <= "00";
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if LoadNShift = '1' then
        ShiftData <= PieceData;
        Player <= PlayerData;
        Cursor <= CursorData;
        Background <= BackgroundData;
      else
        ShiftData <= ShiftData(30 downto 0) & ShiftData(31);
        Player <= Player;
        Cursor <= Cursor;
        Background <= Background;
      end if;
    end if;
  end process ShiftRegister;

  VideoData <= ShiftData(31);

  -- Video signals going to the "video DAC"

  VideoOut: process (Pixel_Clock, OPB_Rst, Background, Player, Cursor)
  begin
    if OPB_Rst = '1' then
      VIDOUT_RED   <= '0';
      VIDOUT_BLUE  <= '0';
      VIDOUT_GREEN <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if HBLANK_N = '1' and VBLANK_N = '1' then
        if VideoData = '1' then
          if Player = '0' and Cursor = '0' then
            VIDOUT_RED   <= '1';
            VIDOUT_GREEN <= '0';
            VIDOUT_BLUE  <= '0';
          elsif Player = '0' and Cursor = '1' then
            VIDOUT_RED   <= '0';
            VIDOUT_GREEN <= '1';
            VIDOUT_BLUE  <= '1';
          elsif Player = '1' and Cursor = '1' then
            VIDOUT_RED   <= '1';
            VIDOUT_GREEN <= '1';
            VIDOUT_BLUE  <= '0';
          else
            VIDOUT_RED   <= '0';
            VIDOUT_GREEN <= '0';
            VIDOUT_BLUE  <= '1';
          end if;
        else
          if Cursor = '1' then
            VIDOUT_RED   <= '0';
            VIDOUT_GREEN <= '1';
            VIDOUT_BLUE  <= '0';
          else
            if Background(1) = '1' then
              VIDOUT_RED   <= '0';
              VIDOUT_GREEN <= '0';
              VIDOUT_BLUE  <= '0';
            elsif Background(0) = '1' then
              VIDOUT_RED   <= '1';
              VIDOUT_GREEN <= '1';
              VIDOUT_BLUE  <= '1';
            else
              if Hcount(0) = '1' xor Vcount(0) = '1' then
                VIDOUT_RED   <= '1';
                VIDOUT_GREEN <= '1';
                VIDOUT_BLUE  <= '1';
              else
                VIDOUT_RED   <= '0';
                VIDOUT_GREEN <= '0';
                VIDOUT_BLUE  <= '0';
              end if;
            end if;
          end if;
        end if;
      else
        VIDOUT_RED   <= '0';
        VIDOUT_GREEN <= '0';
        VIDOUT_BLUE  <= '0';
      end if;
    end if;
  end process VideoOut;

  VIDOUT_CLK <= Pixel_Clock;

end Behavioral;
